#  tests for cis-1.7.3-py36ha15d459_1 (this is a generated file);
print('===== testing package: cis-1.7.3-py36ha15d459_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
from cis.test.runner import run
import os

# Disable multi proccesing completely - the timeout is breaking builds on OS X as well now...
np=0
# Run the tests
run('cis.test.unit', n_processors=np)

# If the test data directory is specified, run the integration tests as well. We disable parallel processing
# of these though as some have side affects, you also have to specify a time-out which would have to be very large
if os.environ.get("CIS_DATA_HOME", None):
    run('cis.test.integration', n_processors=0)
#  --- run_test.py (end) ---

print('===== cis-1.7.3-py36ha15d459_1 OK =====');
print("import: 'cis'")
import cis

