"""
Plotting integration tests. These tests use matplotlib to actually check the image output is 'similar' to some reference
image. The tolerance is defined in the test_plotting script and based onthe default used in iris 1.8.0.
"""